
/*
**
**  $VER: descriptor.c 1.1 (21.7.97)
**  gifanim.datatype 1.1
**
**  DataTypes custom comparision routine for a GIF Animation
**
**  Written 1997 by Roland 'Gizzy' Mainz
**  Original example source from David N. Junod
**
*/

/* main includes */
#include "/classbase.h"

#undef SysBase
#undef DOSBase
#undef UtilityBase
#define SysBase     (dthc -> dthc_SysBase)
#define UtilityBase (dthc -> dthc_UtilityBase)
#define DOSBase     (dthc -> dthc_DOSBase)

#define pm_error kprintf

void __stdargs kprintf( STRPTR, ... );

#if 0
#define D( x ) x
#else
#define D( x )
#endif

#define MAXCOLORMAPSIZE (256)

#define MAX_LWZ_BITS (12)

#define SORTED         (0x20) /* colors sorted, important first */
#define INTERLACE      (0x40) /* interlaced picture             */
#define LOCALCOLORMAP  (0x80) /* frames have their own colormap */
#define BitSet( byte, bit ) (((byte) & (bit)) == (bit))

#define ReadOK( file, buffer, len ) (Read( (file), (buffer), (LONG)(len) ) == (LONG)(len))

#define LM_to_ushrt( a, b ) ((UWORD)(((b) << 8)|(a)))


struct GIFDecoder
{
  struct
  {
    UWORD                Width;
    UWORD                Height;
    struct ColorRegister ColorMap[ MAXCOLORMAPSIZE ]; /* colormap         */
    UWORD                BitPixel;                    /* number of colors */
    UWORD                ColorResolution;
    UBYTE                Background;                  /* background pen   */
    UWORD                AspectRatio;
  } GifScreen;

  struct
  {
    UWORD transparent; /* transparent pen; ~0U for nothing transparent */
    UWORD delayTime;   /* frame delay in 1/100 sec */
    BOOL  inputFlag;   /* user input before continue ? */
    UBYTE disposal;
#define GIF89A_DISPOSE_NOP                  (0U) /* No disposal specified. The decoder is not required to take any action. */
#define GIF89A_DISPOSE_NODISPOSE            (1U) /* Do not dispose. The graphic is to be left in place. */
#define GIF89A_DISPOSE_RESTOREBACKGROUND    (2U) /* Restore to background color. The area used by the graphic must be restored to the background color. */
#define GIF89A_DISPOSE_RESTOREPREVIOUS      (3U) /* Restore to previous. The decoder is required to restore the area overwritten by the graphic with what was there prior to rendering the graphic. */
#define GIF89A_DISPOSE_RESERVED4            (4U)
#define GIF89A_DISPOSE_RESERVED5            (5U)
#define GIF89A_DISPOSE_RESERVED6            (6U)
#define GIF89A_DISPOSE_RESERVED7            (7U)
  } Gif89; /* = { (UWORD)~0U, (UWORD)~0U, FALSE, GIF89A_DISPOSE_NOP };*/
};


DISPATCHERFLAGS
BOOL DTHook( REGA0 struct DTHookContext *dthc )
{
    struct GIFDecoder gifdec = { 0 };
    BOOL              retval = FALSE;
    BPTR              file;

    /* Make sure that this is an IFF data type */
    if( file = dthc -> dthc_FileHandle )
    {
      UBYTE buf[ 16 ];
      UBYTE c;
      BOOL  useGlobalColormap;
      UWORD bitPixel;
      UBYTE version[ 4 ];
      ULONG imagecount = 0UL;

      if( !ReadOK( file, buf, 6 ) )
      {
        D( kprintf( "error reading magic number" ) );
        return( FALSE );
      }

      if( strncmp( (char *)buf, "GIF", 3 ) != 0 )
      {
        D( kprintf( "not a GIF file" ) );
        return( FALSE );
      }

      strncpy( version, (char *)(buf + 3), 3 );
      version[ 3 ] = '\0';

      if( (strcmp( version, "87a" ) != 0) &&
          (strcmp( version, "89a" ) != 0) )
      {
        D( kprintf( "bad version number, not '87a' or '89a'" ) );

        return( FALSE );
      }

      if( !ReadOK( file, buf, 7 ) )
      {
        D( pm_error( "failed to read screen descriptor" ) );
      }

      gifdec . GifScreen . BitPixel        = 2 << (buf[ 4 ] & 0x07);

      /* Global Colormap ? */
      if( BitSet( buf[ 4 ], LOCALCOLORMAP ) )
      {
        /* Skip colormap */
        (void)Seek( file, (long)(sizeof( struct ColorRegister ) * (gifdec . GifScreen . BitPixel)), OFFSET_CURRENT );
      }

      for( ;; )
      {
        if( !ReadOK( file, (&c), 1 ) )
          D( pm_error( "EOF / read error on image data" ) );

        switch( c )
        {
          case ';': /* GIF terminator ? */
              return( FALSE );

          case '!': /* Extension ? */
          {
              UBYTE count;

              if( !ReadOK( file, &c, 1 ) )
                D( pm_error("OF / read error on extention function code") );

              do
              {
                if( !ReadOK( file, &count, 1 ) )
                {
                  D( kprintf( "error in getting DataBlock size\n" ) );

                  return( FALSE );
                }

                if( (count != 0) && (Seek( file, (long)count, OFFSET_CURRENT ) == -1L) )
                {
                  D( kprintf( "error in reading DataBlock\n" ) );

                  return( FALSE );
                }
              } while( count != 0 );
          }
              break;

          case ',': /* Raster data start ? */
          {
              UBYTE c;

              if( imagecount++ > 0UL ) return( TRUE );

              if( !ReadOK( file, buf, 9 ) )
                D( pm_error( "couldn't read left/top/width/height" ) );

              useGlobalColormap = !BitSet( buf[ 8 ], LOCALCOLORMAP );

              bitPixel = 1 << ((buf[ 8 ] & 0x07) + 1);

              if( !useGlobalColormap )
              {
                /* Skip local colormap */
                (void)Seek( file, (long)(sizeof( struct ColorRegister ) * bitPixel), OFFSET_CURRENT );
              }

              /* Initialize the Compression routines */
              if( !ReadOK( file, &c, 1 ) )
                D( pm_error( "EOF / read error on image data" ) );

              D( kprintf( "skipping image...\n" ) );

              /* Loop until end of raster data */
              for( ;; )
              {
                if( !ReadOK( file, &c, 1 ) )
                  D( pm_error( "EOF / reading block byte count" ) );

                if( c == 0 )
                {
                  D( kprintf( "image done\n" ) );
                  break;
                }

                /* Skip... */
                (void)Seek( file, (long)c, OFFSET_CURRENT );
              }
          }
              break;

          default: /* Not a valid raster data start character ? */
          {
              D( kprintf( "invalid character 0x%02x, ignoring\n", (int)c ) );
          }
              break;
        }
      }
    }

    return( retval );
}


